@@ -1,14 +1,12 @@
 /*
  * "lpr" command for CUPS.
  *
- * Copyright 2007-2016 by Apple Inc.
- * Copyright 1997-2007 by Easy Software Products.
+ * Copyright © 2021 by OpenPrinting.
+ * Copyright © 2007-2019 by Apple Inc.
+ * Copyright © 1997-2007 by Easy Software Products.
  *
- * These coded instructions, statements, and computer programs are the
- * property of Apple Inc. and are protected by Federal copyright
- * law.  Distribution and use rights are outlined in the file "LICENSE.txt"
- * which should have been included with this file.  If this file is
- * missing or damaged, see the license at "http://www.cups.org/".
+ * Licensed under Apache License v2.0.  See the file "LICENSE" for more
+ * information.
  */
 
 /*
@@ -18,6 +16,13 @@
 #include <cups/cups-private.h>
 
 
+/*
+ * Local functions...
+ */
+
+static void	usage(void) _CUPS_NORETURN;
+
+
 /*
  * 'main()' - Parse options and send files for printing.
  */
@@ -55,18 +60,20 @@ main(int  argc,				/* I - Number of command-line arguments */
 
   for (i = 1; i < argc; i ++)
   {
-    if (argv[i][0] == '-')
+    if (!strcmp(argv[i], "--help"))
+      usage();
+    else if (argv[i][0] == '-')
     {
       for (opt = argv[i] + 1; *opt; opt ++)
       {
 	switch (ch = *opt)
 	{
 	  case 'E' : /* Encrypt */
-#ifdef HAVE_SSL
+#ifdef HAVE_TLS
 	      cupsSetEncryption(HTTP_ENCRYPT_REQUIRED);
 #else
 	      _cupsLangPrintf(stderr, _("%s: Sorry, no encryption support."), argv[0]);
-#endif /* HAVE_SSL */
+#endif /* HAVE_TLS */
 	      break;
 
 	  case 'U' : /* Username */
@@ -81,7 +88,7 @@ main(int  argc,				/* I - Number of command-line arguments */
 		if (i >= argc)
 		{
 		  _cupsLangPrintf(stderr, _("%s: Error - expected username after \"-U\" option."), argv[0]);
-		  return (1);
+		  usage();
 		}
 
 		cupsSetUser(argv[i]);
@@ -101,7 +108,7 @@ main(int  argc,				/* I - Number of command-line arguments */
 		if (i >= argc)
 		{
 		  _cupsLangPrintf(stderr, _("%s: Error - expected hostname after \"-H\" option."), argv[0]);
-		  return (1);
+		  usage();
 		}
 		else
 		  cupsSetServer(argv[i]);
@@ -127,7 +134,7 @@ main(int  argc,				/* I - Number of command-line arguments */
 		  _cupsLangPrintf(stderr,
 				  _("%s: Error - expected value after \"-%c\" "
 				    "option."), argv[0], ch);
-		  return (1);
+		  usage();
 		}
 	      }
 
@@ -153,7 +160,7 @@ main(int  argc,				/* I - Number of command-line arguments */
 		if (i >= argc)
 		{
 		  _cupsLangPrintf(stderr, _("%s: Error - expected option=value after \"-o\" option."), argv[0]);
-		  return (1);
+		  usage();
 		}
 
 		num_options = cupsParseOptions(argv[i], num_options, &options);
@@ -204,7 +211,7 @@ main(int  argc,				/* I - Number of command-line arguments */
 		if (i >= argc)
 		{
 		  _cupsLangPrintf(stderr, _("%s: Error - expected destination after \"-P\" option."), argv[0]);
-		  return (1);
+		  usage();
 		}
 
 		printer = argv[i];
@@ -228,6 +235,12 @@ main(int  argc,				/* I - Number of command-line arguments */
 		_cupsLangPrintf(stderr, _("%s: Error - add '/version=1.1' to server name."), argv[0]);
 		return (1);
 	      }
+	      else if (cupsLastError() == IPP_STATUS_ERROR_NOT_FOUND)
+	      {
+		_cupsLangPrintf(stderr,
+				_("%s: Error - The printer or class does not exist."), argv[0]);
+		return (1);
+	      }
 	      break;
 
 	  case '#' : /* Number of copies */
@@ -242,7 +255,7 @@ main(int  argc,				/* I - Number of command-line arguments */
 		if (i >= argc)
 		{
 		  _cupsLangPrintf(stderr, _("%s: Error - expected copies after \"-#\" option."), argv[0]);
-		  return (1);
+		  usage();
 		}
 
 		num_copies = atoi(argv[i]);
@@ -254,8 +267,7 @@ main(int  argc,				/* I - Number of command-line arguments */
 		return (1);
 	      }
 
-	      sprintf(buffer, "%d", num_copies);
-	      num_options = cupsAddOption("copies", buffer, num_options, &options);
+	      num_options = cupsAddIntegerOption("copies", num_copies, num_options, &options);
 	      break;
 
 	  case 'C' : /* Class */
@@ -272,7 +284,7 @@ main(int  argc,				/* I - Number of command-line arguments */
 		if (i >= argc)
 		{
 		  _cupsLangPrintf(stderr, _("%s: Error - expected name after \"-%c\" option."), argv[0], ch);
-		  return (1);
+		  usage();
 		}
 
 		title = argv[i];
@@ -412,3 +424,38 @@ main(int  argc,				/* I - Number of command-line arguments */
 
   return (0);
 }
+
+
+/*
+ * 'usage()' - Show program usage and exit.
+ */
+
+static void
+usage(void)
+{
+  _cupsLangPuts(stdout, _("Usage: lpr [options] [file(s)]"));
+  _cupsLangPuts(stdout, _("Options:"));
+  _cupsLangPuts(stdout, _("-# num-copies           Specify the number of copies to print"));
+  _cupsLangPuts(stdout, _("-E                      Encrypt the connection to the server"));
+  _cupsLangPuts(stdout, _("-H server[:port]        Connect to the named server and port"));
+  _cupsLangPuts(stdout, _("-m                      Send an email notification when the job completes"));
+  _cupsLangPuts(stdout, _("-o option[=value]       Specify a printer-specific option"));
+  _cupsLangPuts(stdout, _("-o job-sheets=standard  Print a banner page with the job"));
+  _cupsLangPuts(stdout, _("-o media=size           Specify the media size to use"));
+  _cupsLangPuts(stdout, _("-o number-up=N          Specify that input pages should be printed N-up (1, 2, 4, 6, 9, and 16 are supported)"));
+  _cupsLangPuts(stdout, _("-o orientation-requested=N\n"
+                          "                        Specify portrait (3) or landscape (4) orientation"));
+  _cupsLangPuts(stdout, _("-o print-quality=N      Specify the print quality - draft (3), normal (4), or best (5)"));
+  _cupsLangPuts(stdout, _("-o sides=one-sided      Specify 1-sided printing"));
+  _cupsLangPuts(stdout, _("-o sides=two-sided-long-edge\n"
+                          "                        Specify 2-sided portrait printing"));
+  _cupsLangPuts(stdout, _("-o sides=two-sided-short-edge\n"
+                          "                        Specify 2-sided landscape printing"));
+  _cupsLangPuts(stdout, _("-P destination          Specify the destination"));
+  _cupsLangPuts(stdout, _("-q                      Specify the job should be held for printing"));
+  _cupsLangPuts(stdout, _("-r                      Remove the file(s) after submission"));
+  _cupsLangPuts(stdout, _("-T title                Specify the job title"));
+  _cupsLangPuts(stdout, _("-U username             Specify the username to use for authentication"));
+
+  exit(1);
+}
