@@ -1,16 +1,12 @@
 /*
  * HTTP test program for CUPS.
  *
+ * Copyright © 2021 by OpenPrinting.
  * Copyright © 2007-2018 by Apple Inc.
  * Copyright © 1997-2006 by Easy Software Products.
  *
- * These coded instructions, statements, and computer programs are the
- * property of Apple Inc. and are protected by Federal copyright
- * law.  Distribution and use rights are outlined in the file "LICENSE.txt"
- * which should have been included with this file.  If this file is
- * missing or damaged, see the license at "http://www.cups.org/".
- *
- * This file is subject to the Apple OS-Developed Software exception.
+ * Licensed under Apache License v2.0.  See the file "LICENSE" for more
+ * information.
  */
 
 /*
@@ -785,7 +781,7 @@ main(int  argc,				/* I - Number of command-line arguments */
 
 	continue;
       }
-#ifdef HAVE_SSL
+#ifdef HAVE_TLS
       else if (status == HTTP_STATUS_UPGRADE_REQUIRED)
       {
 	/* Flush any error message... */
@@ -804,7 +800,7 @@ main(int  argc,				/* I - Number of command-line arguments */
 	/* Try again, this time with encryption enabled... */
 	continue;
       }
-#endif /* HAVE_SSL */
+#endif /* HAVE_TLS */
     }
     while (status == HTTP_STATUS_UNAUTHORIZED ||
            status == HTTP_STATUS_UPGRADE_REQUIRED);
@@ -887,7 +883,7 @@ main(int  argc,				/* I - Number of command-line arguments */
 
 	continue;
       }
-#ifdef HAVE_SSL
+#ifdef HAVE_TLS
       else if (status == HTTP_STATUS_UPGRADE_REQUIRED)
       {
 	/* Flush any error message... */
@@ -906,7 +902,7 @@ main(int  argc,				/* I - Number of command-line arguments */
 	/* Try again, this time with encryption enabled... */
 	continue;
       }
-#endif /* HAVE_SSL */
+#endif /* HAVE_TLS */
     }
     while (status == HTTP_STATUS_UNAUTHORIZED || status == HTTP_STATUS_UPGRADE_REQUIRED);
 
