@@ -1,17 +1,15 @@
 /*
  * Authorization routines for the CUPS scheduler.
  *
- * Copyright 2007-2016 by Apple Inc.
- * Copyright 1997-2007 by Easy Software Products, all rights reserved.
+ * Copyright © 2021 by OpenPrinting.
+ * Copyright © 2007-2019 by Apple Inc.
+ * Copyright © 1997-2007 by Easy Software Products, all rights reserved.
  *
  * This file contains Kerberos support code, copyright 2006 by
  * Jelmer Vernooij.
  *
- * These coded instructions, statements, and computer programs are the
- * property of Apple Inc. and are protected by Federal copyright
- * law.  Distribution and use rights are outlined in the file "LICENSE.txt"
- * which should have been included with this file.  If this file is
- * missing or damaged, see the license at "http://www.cups.org/".
+ * Licensed under Apache License v2.0.  See the file "LICENSE" for more
+ * information.
  */
 
 /*
@@ -38,11 +36,6 @@
 #endif /* HAVE_MEMBERSHIP_H */
 #ifdef HAVE_AUTHORIZATION_H
 #  include <Security/AuthorizationTags.h>
-#  ifdef HAVE_SECBASEPRIV_H
-#    include <Security/SecBasePriv.h>
-#  else
-extern const char *cssmErrorString(int error);
-#  endif /* HAVE_SECBASEPRIV_H */
 #endif /* HAVE_AUTHORIZATION_H */
 #ifdef HAVE_SYS_PARAM_H
 #  include <sys/param.h>
@@ -59,27 +52,30 @@ typedef struct sockpeercred cupsd_ucred_t;
 #  endif
 #  define CUPSD_UCRED_UID(c) (c).uid
 #endif /* HAVE_SYS_UCRED_H */
+#ifdef HAVE_LIBAPPARMOR
+#  include <sys/apparmor.h>
+#endif /* HAVE_LIBAPPARMOR */
+#ifdef HAVE_LIBSNAPDGLIB
+#  include <glib.h>
+#  include <snapd-glib/snapd-glib.h>
+#endif /* HAVE_LIBSNAPDGLIB */
 
 
 /*
  * Local functions...
  */
 
+static int		check_admin_access(cupsd_client_t *con);
 #ifdef HAVE_AUTHORIZATION_H
 static int		check_authref(cupsd_client_t *con, const char *right);
 #endif /* HAVE_AUTHORIZATION_H */
 static int		compare_locations(cupsd_location_t *a,
 			                  cupsd_location_t *b);
 static cupsd_authmask_t	*copy_authmask(cupsd_authmask_t *am, void *data);
-#if !HAVE_LIBPAM
-static char		*cups_crypt(const char *pw, const char *salt);
-#endif /* !HAVE_LIBPAM */
 static void		free_authmask(cupsd_authmask_t *am, void *data);
 #if HAVE_LIBPAM
 static int		pam_func(int, const struct pam_message **,
 			         struct pam_response **, void *);
-#else
-static void		to64(char *s, unsigned long v, int n);
 #endif /* HAVE_LIBPAM */
 
 
@@ -215,7 +211,6 @@ cupsdAddNameMask(cups_array_t **masks,	/* IO - Masks array (created as needed) *
 
     if (ifptr >= ifname && *ifptr == ')')
     {
-      ifptr --;
       *ifptr = '\0';
     }
 
@@ -350,7 +345,7 @@ cupsdAuthorize(cupsd_client_t *con)	/* I - Client connection */
 
     if ((status = AuthorizationCreateFromExternalForm((AuthorizationExternalForm *)authdata, &con->authref)) != 0)
     {
-      cupsdLogClient(con, CUPSD_LOG_ERROR, "AuthorizationCreateFromExternalForm returned %d (%s)", (int)status, cssmErrorString(status));
+      cupsdLogClient(con, CUPSD_LOG_ERROR, "AuthorizationCreateFromExternalForm returned %d", (int)status);
       return;
     }
 
@@ -694,14 +689,14 @@ cupsdAuthorize(cupsd_client_t *con)	/* I - Client connection */
 	    * client...
 	    */
 
-	    pass = cups_crypt(password, pw->pw_passwd);
+	    pass = crypt(password, pw->pw_passwd);
 
 	    if (!pass || strcmp(pw->pw_passwd, pass))
 	    {
 #  ifdef HAVE_SHADOW_H
 	      if (spw)
 	      {
-		pass = cups_crypt(password, spw->sp_pwdp);
+		pass = crypt(password, spw->sp_pwdp);
 
 		if (pass == NULL || strcmp(spw->sp_pwdp, pass))
 		{
@@ -738,7 +733,6 @@ cupsdAuthorize(cupsd_client_t *con)	/* I - Client connection */
 					/* Output token for username */
     gss_name_t		client_name;	/* Client name */
 
-
 #  ifdef __APPLE__
    /*
     * If the weak-linked GSSAPI/Kerberos library is not present, don't try
@@ -770,7 +764,7 @@ cupsdAuthorize(cupsd_client_t *con)	/* I - Client connection */
     * Decode the authorization string to get the input token...
     */
 
-    len                = (int)strlen(authorization);
+    len                = (int)strlen(authorization) + 0;
     input_token.value  = malloc((size_t)len);
     input_token.value  = httpDecode64_2(input_token.value, &len,
 					authorization);
@@ -1179,7 +1173,23 @@ cupsdCheckGroup(
 
     groupid = group->gr_gid;
 
+    for (i = 0; group->gr_mem[i]; i ++)
+    {
+     /*
+      * User appears in the group membership...
+      */
+
+      if (!_cups_strcasecmp(username, group->gr_mem[i]))
+	return (1);
+    }
+
 #ifdef HAVE_GETGROUPLIST
+   /*
+    * If the user isn't in the group membership list, try the results from
+    * getgrouplist() which is supposed to return the full list of groups a user
+    * belongs to...
+    */
+
     if (user)
     {
       int	ngroups;		/* Number of groups */
@@ -1200,13 +1210,6 @@ cupsdCheckGroup(
         if ((int)groupid == (int)groups[i])
 	  return (1);
     }
-
-#else
-    for (i = 0; group->gr_mem[i]; i ++)
-    {
-      if (!_cups_strcasecmp(username, group->gr_mem[i]))
-	return (1);
-    }
 #endif /* HAVE_GETGROUPLIST */
   }
   else
@@ -1633,7 +1636,7 @@ cupsdIsAuthorized(cupsd_client_t *con,	/* I - Connection */
   if (auth == CUPSD_AUTH_DENY && best->satisfy == CUPSD_AUTH_SATISFY_ALL)
     return (HTTP_FORBIDDEN);
 
-#ifdef HAVE_SSL
+#ifdef HAVE_TLS
  /*
   * See if encryption is required...
   */
@@ -1650,7 +1653,7 @@ cupsdIsAuthorized(cupsd_client_t *con,	/* I - Connection */
                     "cupsdIsAuthorized: Need upgrade to TLS...");
     return (HTTP_UPGRADE_REQUIRED);
   }
-#endif /* HAVE_SSL */
+#endif /* HAVE_TLS */
 
  /*
   * Now see what access level is required...
@@ -1718,12 +1721,9 @@ cupsdIsAuthorized(cupsd_client_t *con,	/* I - Connection */
 
  /*
   * OK, got a username.  See if we need normal user access, or group
-  * access... (root always matches)
+  * access...
   */
 
-  if (!strcmp(username, "root"))
-    return (HTTP_OK);
-
  /*
   * Strip any @domain or @KDC from the username and owner...
   */
@@ -1753,6 +1753,21 @@ cupsdIsAuthorized(cupsd_client_t *con,	/* I - Connection */
   else
     pw = NULL;
 
+ /*
+  * For matching user and group memberships below we will first go
+  * through all names except @SYSTEM to authorize the task as
+  * non-administrative, like printing or deleting one's own job, if this
+  * fails we will check whether we can authorize via the special name
+  * @SYSTEM, as an administrative task, like creating a print queue or
+  * deleting someone else's job.
+  * Note that tasks are considered as administrative by the policies
+  * in cupsd.conf, when they require the user or group @SYSTEM.
+  * We do this separation because if the client is a Snap connecting via
+  * domain socket, we need to additionally check whether it plugs to us
+  * through the "cups-control" interface which allows administration and
+  * not through the "cups" interface which allows only printing.
+  */
+
   if (best->level == CUPSD_AUTH_USER)
   {
    /*
@@ -1783,8 +1798,14 @@ cupsdIsAuthorized(cupsd_client_t *con,	/* I - Connection */
       {
 	if (!_cups_strncasecmp(name, "@AUTHKEY(", 9) && check_authref(con, name + 9))
 	  return (HTTP_OK);
-	else if (!_cups_strcasecmp(name, "@SYSTEM") && SystemGroupAuthKey &&
-		 check_authref(con, SystemGroupAuthKey))
+      }
+
+      for (name = (char *)cupsArrayFirst(best->names);
+           name;
+	   name = (char *)cupsArrayNext(best->names))
+      {
+	if (!_cups_strcasecmp(name, "@SYSTEM") && SystemGroupAuthKey &&
+	    check_authref(con, SystemGroupAuthKey))
 	  return (HTTP_OK);
       }
 
@@ -1801,9 +1822,8 @@ cupsdIsAuthorized(cupsd_client_t *con,	/* I - Connection */
 	return (HTTP_OK);
       else if (!_cups_strcasecmp(name, "@SYSTEM"))
       {
-        for (i = 0; i < NumSystemGroups; i ++)
-	  if (cupsdCheckGroup(username, pw, SystemGroups[i]))
-	    return (HTTP_OK);
+	/* Do @SYSTEM later, when every other entry fails */
+	continue;
       }
       else if (name[0] == '@')
       {
@@ -1814,6 +1834,18 @@ cupsdIsAuthorized(cupsd_client_t *con,	/* I - Connection */
         return (HTTP_OK);
     }
 
+    for (name = (char *)cupsArrayFirst(best->names);
+	 name;
+	 name = (char *)cupsArrayNext(best->names))
+    {
+      if (!_cups_strcasecmp(name, "@SYSTEM"))
+      {
+        for (i = 0; i < NumSystemGroups; i ++)
+	  if (cupsdCheckGroup(username, pw, SystemGroups[i]) && check_admin_access(con))
+	    return (HTTP_OK);
+      }
+    }
+
     return (con->username[0] ? HTTP_FORBIDDEN : HTTP_UNAUTHORIZED);
   }
 
@@ -1831,16 +1863,30 @@ cupsdIsAuthorized(cupsd_client_t *con,	/* I - Connection */
        name;
        name = (char *)cupsArrayNext(best->names))
   {
-    cupsdLogMessage(CUPSD_LOG_DEBUG2, "cupsdIsAuthorized: Checking group \"%s\" membership...", name);
-
     if (!_cups_strcasecmp(name, "@SYSTEM"))
     {
+      /* Do @SYSTEM later, when every other entry fails */
+      continue;
+    }
+
+    cupsdLogMessage(CUPSD_LOG_DEBUG2, "cupsdIsAuthorized: Checking group \"%s\" membership...", name);
+
+    if (cupsdCheckGroup(username, pw, name))
+      return (HTTP_OK);
+  }
+
+  for (name = (char *)cupsArrayFirst(best->names);
+       name;
+       name = (char *)cupsArrayNext(best->names))
+  {
+    if (!_cups_strcasecmp(name, "@SYSTEM"))
+    {
+      cupsdLogMessage(CUPSD_LOG_DEBUG2, "cupsdIsAuthorized: Checking group \"%s\" membership...", name);
+
       for (i = 0; i < NumSystemGroups; i ++)
-	if (cupsdCheckGroup(username, pw, SystemGroups[i]))
+	if (cupsdCheckGroup(username, pw, SystemGroups[i]) && check_admin_access(con))
 	  return (HTTP_OK);
     }
-    else if (cupsdCheckGroup(username, pw, name))
-      return (HTTP_OK);
   }
 
  /*
@@ -1893,6 +1939,219 @@ cupsdNewLocation(const char *location)	/* I - Location path */
 }
 
 
+/*
+ * 'check_admin_access()' - Verify that the client has administrative access.
+ */
+
+static int				// O - 1 if authorized, 0 otherwise
+check_admin_access(cupsd_client_t *con) // I - Client connection
+{
+#if defined(HAVE_LIBAPPARMOR) && defined(HAVE_LIBSNAPDGLIB)
+ /*
+  * If the client accesses locally via domain socket, find out whether it
+  * is a Snap.  Grant access if it is not a Snap, if it is a classic Snap
+  * or if it is a confined Snap which plugs "cups-control".  Otherwise deny
+  * access.
+  */
+
+  int		fd = httpGetFd(con->http);
+					// Client socket file descriptor
+  char		*context = NULL;	// AppArmor profile name of client
+  SnapdClient	*client = NULL;		// Data structure of snapd access
+  GError	*error = NULL;		// Glib error
+  int		ret = 1;		// Return value
+#  if !CUPS_SNAP
+  SnapdSnap	*snap = NULL;		// Data structure of client Snap
+#  endif // !CUPS_SNAP
+
+
+#  ifdef AF_LOCAL
+  // Only check domain sockets...
+  if (httpAddrFamily(con->http->hostaddr) != AF_LOCAL)
+    return (1);
+#  endif // AF_LOCAL
+
+#  if !CUPS_SNAP
+  // If AppArmor is not enabled, then we can't identify the client...
+  if (!aa_is_enabled())
+  {
+    cupsdLogClient(con, CUPSD_LOG_DEBUG, "AppArmor not in use.");
+    return (1);
+  }
+#  endif /* !CUPS_SNAP */
+
+  // Get the client's AppArmor context using the socket...
+  if (aa_getpeercon(fd, &context, NULL) < 0)
+  {
+    cupsdLogClient(con, CUPSD_LOG_DEBUG, "AppArmor profile could not be retrieved: %s", strerror(errno));
+    return (1);
+  }
+  else
+  {
+    cupsdLogClient(con, CUPSD_LOG_DEBUG, "AppArmor profile is '%s'.", context);
+  }
+
+  // Allow access from "cups" snap...
+  if (!strncmp(context, "snap.cups.", 10))
+  {
+    cupsdLogClient(con, CUPSD_LOG_DEBUG, "Client from the CUPS Snap itself - allowed.");
+    goto done;
+  }
+
+#  if CUPS_SNAP && defined(HAVE_SNAPD_CLIENT_RUN_SNAPCTL2_SYNC)
+ /*
+  * CUPS is snapped, so check whether the client is also snapped.  If so,
+  * determine whether the client snap has a "cups-control" plug which allows
+  * the application to perform CUPS administrative tasks.
+  */
+
+  const char	*cookie;		// snapd access cookie
+  int		status = 65535;		// Status of client Snap context check
+  const char	*args[] =		// snapctl arguments
+  {
+    "is-connected",
+    "--apparmor-label",
+    NULL,
+    "cups-control",
+    NULL
+  };
+
+  // Connect to snapd
+  if ((client = snapd_client_new()) == NULL)
+  {
+    cupsdLogClient(con, CUPSD_LOG_ERROR, "Unable to connect to snapd.");
+    ret = 0;
+    goto done;
+  }
+
+  // snapctl commands are sent over a domain socket
+  snapd_client_set_socket_path(client, "/run/snapd-snap.socket");
+
+  // Take cookie from the environment if available
+  if ((cookie = g_getenv("SNAP_COOKIE")) == NULL)
+  {
+    cookie = "";
+    cupsdLogClient(con, CUPSD_LOG_WARN, "No SNAP_COOKIE set in the Snap environment.");
+  }
+
+  // Do the client Snap context check...
+  args[2] = context;
+
+  if (!snapd_client_run_snapctl2_sync(client, cookie, (char **)args, NULL, NULL, &status, NULL, &error))
+  {
+    cupsdLogClient(con, CUPSD_LOG_ERROR, "Unable to check snap context: %s", error->message);
+    ret = 0;
+    goto done;
+  }
+
+  switch (status)
+  {
+    case 0 : // The client is a confined Snap and plugs cups-control
+	cupsdLogClient(con, CUPSD_LOG_DEBUG, "Snap with cups-control plug - allowed.");
+	break;
+    case 1 : // The client is a confined Snap and does not plug cups-control
+	cupsdLogClient(con, CUPSD_LOG_DEBUG, "Snap without cups-control plug - denied.");
+	ret = 0;
+	break;
+    case 10 : // The client is a classic Snap
+	cupsdLogClient(con, CUPSD_LOG_DEBUG, "Classic snap - allowed.");
+	break;
+    case 11 : // The client is not a Snap
+	cupsdLogClient(con, CUPSD_LOG_DEBUG, "Not a snap - allowed.");
+	break;
+    default : // Unexpected status...
+	cupsdLogClient(con, CUPSD_LOG_ERROR, "Snap check returned unexpected status %d - denied.", status);
+	ret = 0;
+	break;
+  }
+
+#  elif !CUPS_SNAP
+ /*
+  * If CUPS is not snapped, check whether the client is snapped and if it has
+  * the "cups-control" plug.
+  */
+
+  // Is the client a snapped application?
+  if (strncmp(context, "snap.", 5))
+  {
+    cupsdLogClient(con, CUPSD_LOG_DEBUG, "Not a snap - allowed.");
+    goto done;
+  }
+
+  // Extract the snap name from the context (snap.name.instance)
+  char *snap_name = strdup(context + 5);// Snap name follows "snap."
+  char *ptr = strchr(snap_name, '.');	// instance follows the name...
+  if (!ptr)
+  {
+    cupsdLogClient(con, CUPSD_LOG_DEBUG, "Malformed snapd AppArmor profile name '%s' - denied.", context);
+    free(snap_name);
+    ret = 0;
+    goto done;
+  }
+
+  *ptr = '\0';
+  cupsdLogClient(con, CUPSD_LOG_DEBUG, "Client snap is '%s'.", snap_name);
+
+  // Connect to snapd
+  if ((client = snapd_client_new()) == NULL)
+  {
+    cupsdLogClient(con, CUPSD_LOG_ERROR, "Unable to connect to snapd.");
+    free(snap_name);
+    ret = 0;
+    goto done;
+  }
+
+  // Check whether the client Snap is under classic confinement
+  GPtrArray *plugs = NULL;		// List of plugs for snap
+
+  if ((snap = snapd_client_get_snap_sync(client, snap_name, NULL, &error)) == NULL)
+  {
+    cupsdLogClient(con, CUPSD_LOG_DEBUG, "Unable to get client Snap data: %s", error->message);
+    ret = 0;
+  }
+  // Snaps using classic confinement are granted access
+  else if (snapd_snap_get_confinement(snap) == SNAPD_CONFINEMENT_CLASSIC)
+  {
+    cupsdLogClient(con, CUPSD_LOG_DEBUG, "Classic snap - allowed.");
+  }
+  // Check whether the client Snap has the cups-control plug
+  else if (!snapd_client_get_connections2_sync(client, SNAPD_GET_CONNECTIONS_FLAGS_NONE, snap_name, "cups-control", NULL, NULL, &plugs, NULL, NULL, &error))
+  {
+    cupsdLogClient(con, CUPSD_LOG_DEBUG, "Unable to get client Snap plugs: %s", error->message);
+    ret = 0;
+  }
+  else if (!plugs || plugs->len <= 0)
+  {
+    cupsdLogClient(con, CUPSD_LOG_DEBUG, "Snap without cups-control plug - denied.");
+    ret = 0;
+  }
+  else
+  {
+    cupsdLogClient(con, CUPSD_LOG_DEBUG, "Snap with cups-control plug - allowed.");
+  }
+
+  if (plugs)
+    g_ptr_array_unref(plugs);
+
+  free(snap_name);
+  g_clear_object(&snap);
+
+#  endif // CUPS_SNAP
+
+  done:
+
+  free(context);
+  g_clear_object(&client);
+
+  return (ret);
+
+#else
+  // No AppArmor/snapd to deal with...
+  return (1);
+#endif // HAVE_LIBAPPARMOR && HAVE_LIBSNAPDGLIB
+}
+
+
 #ifdef HAVE_AUTHORIZATION_H
 /*
  * 'check_authref()' - Check if an authorization services reference has the
@@ -1931,9 +2190,7 @@ check_authref(cupsd_client_t *con,	/* I - Connection */
 					kAuthorizationEmptyEnvironment,
 					authflags, NULL)) != 0)
   {
-    cupsdLogMessage(CUPSD_LOG_ERROR,
-		    "AuthorizationCopyRights(\"%s\") returned %d (%s)",
-		    authright.name, (int)status, cssmErrorString(status));
+    cupsdLogMessage(CUPSD_LOG_ERROR, "AuthorizationCopyRights(\"%s\") returned %d", authright.name, (int)status);
     return (0);
   }
 
@@ -1995,129 +2252,6 @@ copy_authmask(cupsd_authmask_t *mask,	/* I - Existing auth mask */
 }
 
 
-#if !HAVE_LIBPAM
-/*
- * 'cups_crypt()' - Encrypt the password using the DES or MD5 algorithms,
- *                  as needed.
- */
-
-static char *				/* O - Encrypted password */
-cups_crypt(const char *pw,		/* I - Password string */
-           const char *salt)		/* I - Salt (key) string */
-{
-  if (!strncmp(salt, "$1$", 3))
-  {
-   /*
-    * Use MD5 passwords without the benefit of PAM; this is for
-    * Slackware Linux, and the algorithm was taken from the
-    * old shadow-19990827/lib/md5crypt.c source code... :(
-    */
-
-    int			i;		/* Looping var */
-    unsigned long	n;		/* Output number */
-    int			pwlen;		/* Length of password string */
-    const char		*salt_end;	/* End of "salt" data for MD5 */
-    char		*ptr;		/* Pointer into result string */
-    _cups_md5_state_t	state;		/* Primary MD5 state info */
-    _cups_md5_state_t	state2;		/* Secondary MD5 state info */
-    unsigned char	digest[16];	/* MD5 digest result */
-    static char		result[120];	/* Final password string */
-
-
-   /*
-    * Get the salt data between dollar signs, e.g. $1$saltdata$md5.
-    * Get a maximum of 8 characters of salt data after $1$...
-    */
-
-    for (salt_end = salt + 3; *salt_end && (salt_end - salt) < 11; salt_end ++)
-      if (*salt_end == '$')
-        break;
-
-   /*
-    * Compute the MD5 sum we need...
-    */
-
-    pwlen = strlen(pw);
-
-    _cupsMD5Init(&state);
-    _cupsMD5Append(&state, (unsigned char *)pw, pwlen);
-    _cupsMD5Append(&state, (unsigned char *)salt, salt_end - salt);
-
-    _cupsMD5Init(&state2);
-    _cupsMD5Append(&state2, (unsigned char *)pw, pwlen);
-    _cupsMD5Append(&state2, (unsigned char *)salt + 3, salt_end - salt - 3);
-    _cupsMD5Append(&state2, (unsigned char *)pw, pwlen);
-    _cupsMD5Finish(&state2, digest);
-
-    for (i = pwlen; i > 0; i -= 16)
-      _cupsMD5Append(&state, digest, i > 16 ? 16 : i);
-
-    for (i = pwlen; i > 0; i >>= 1)
-      _cupsMD5Append(&state, (unsigned char *)((i & 1) ? "" : pw), 1);
-
-    _cupsMD5Finish(&state, digest);
-
-    for (i = 0; i < 1000; i ++)
-    {
-      _cupsMD5Init(&state);
-
-      if (i & 1)
-        _cupsMD5Append(&state, (unsigned char *)pw, pwlen);
-      else
-        _cupsMD5Append(&state, digest, 16);
-
-      if (i % 3)
-        _cupsMD5Append(&state, (unsigned char *)salt + 3, salt_end - salt - 3);
-
-      if (i % 7)
-        _cupsMD5Append(&state, (unsigned char *)pw, pwlen);
-
-      if (i & 1)
-        _cupsMD5Append(&state, digest, 16);
-      else
-        _cupsMD5Append(&state, (unsigned char *)pw, pwlen);
-
-      _cupsMD5Finish(&state, digest);
-    }
-
-   /*
-    * Copy the final sum to the result string and return...
-    */
-
-    memcpy(result, salt, (size_t)(salt_end - salt));
-    ptr = result + (salt_end - salt);
-    *ptr++ = '$';
-
-    for (i = 0; i < 5; i ++, ptr += 4)
-    {
-      n = ((((unsigned)digest[i] << 8) | (unsigned)digest[i + 6]) << 8);
-
-      if (i < 4)
-        n |= (unsigned)digest[i + 12];
-      else
-        n |= (unsigned)digest[5];
-
-      to64(ptr, n, 4);
-    }
-
-    to64(ptr, (unsigned)digest[11], 2);
-    ptr += 2;
-    *ptr = '\0';
-
-    return (result);
-  }
-  else
-  {
-   /*
-    * Use the standard crypt() function...
-    */
-
-    return (crypt(pw, salt));
-  }
-}
-#endif /* !HAVE_LIBPAM */
-
-
 /*
  * 'free_authmask()' - Free function for auth masks.
  */
@@ -2164,45 +2298,33 @@ pam_func(
   * Answer all of the messages...
   */
 
-  DEBUG_printf(("pam_func: appdata_ptr = %p\n", appdata_ptr));
-
   data = (cupsd_authdata_t *)appdata_ptr;
 
   for (i = 0; i < num_msg; i ++)
   {
-    DEBUG_printf(("pam_func: Message = \"%s\"\n", msg[i]->msg));
-
     switch (msg[i]->msg_style)
     {
       case PAM_PROMPT_ECHO_ON:
-          DEBUG_printf(("pam_func: PAM_PROMPT_ECHO_ON, returning \"%s\"...\n",
-	                data->username));
           replies[i].resp_retcode = PAM_SUCCESS;
           replies[i].resp         = strdup(data->username);
           break;
 
       case PAM_PROMPT_ECHO_OFF:
-          DEBUG_printf(("pam_func: PAM_PROMPT_ECHO_OFF, returning \"%s\"...\n",
-	                data->password));
           replies[i].resp_retcode = PAM_SUCCESS;
           replies[i].resp         = strdup(data->password);
           break;
 
       case PAM_TEXT_INFO:
-          DEBUG_puts("pam_func: PAM_TEXT_INFO...");
           replies[i].resp_retcode = PAM_SUCCESS;
           replies[i].resp         = NULL;
           break;
 
       case PAM_ERROR_MSG:
-          DEBUG_puts("pam_func: PAM_ERROR_MSG...");
           replies[i].resp_retcode = PAM_SUCCESS;
           replies[i].resp         = NULL;
           break;
 
       default:
-          DEBUG_printf(("pam_func: Unknown PAM message %d...\n",
-	                msg[i]->msg_style));
           free(replies);
           return (PAM_CONV_ERR);
     }
@@ -2216,24 +2338,4 @@ pam_func(
 
   return (PAM_SUCCESS);
 }
-#else
-
-
-/*
- * 'to64()' - Base64-encode an integer value...
- */
-
-static void
-to64(char          *s,			/* O - Output string */
-     unsigned long v,			/* I - Value to encode */
-     int           n)			/* I - Number of digits */
-{
-  const char	*itoa64 = "./0123456789"
-                          "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
-                          "abcdefghijklmnopqrstuvwxyz";
-
-
-  for (; n > 0; n --, v >>= 6)
-    *s++ = itoa64[v & 0x3f];
-}
 #endif /* HAVE_LIBPAM */
